/***************************************************************************\
  "table.h" - Symbol Table (Phase C)
  Revision 10; 10 Nov 2002 - 20 Jan 2003
  Copyright (C) 2002-2003  Kevin P. Albrecht

  References:
    Dragon pp433-437
    MCD 2.1.11; 2.1.11.3
\***************************************************************************/

#ifndef TABLE_H
#define TABLE_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //
// GLOBAL_VALUE_STRING_SIZE

// This value (known as "m") should be a prime number so that
//   if "n" is the average number of symbols in the symbol table,
//   "n / m" is approximately <= 2.  See (Aho pp433-437).
// According to Grune, a good "m" is 1008 (MCD 2.1.11.3)
#define TABLE_HASH_ARRAY_SIZE 17

// This value (known as "alpha") is multiplied by each character
//   in the symbol's key before they are added together.
//   According to (Aho p437), alpha=1 is the least efficient
//   and alpha=65599 is best for 32-bit integers.
// According to Grune, a good "alpha" is 613 (MCD 2.1.11.3)
#define TABLE_HASH_ALPHA 65599

#define TABLE_MAX_SYMBOLS 256
#define ROUTINE_MAX_PARAMETERS 32

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "global.h"
#include "token.h"
#include "global_conv.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

typedef enum
{
  tNull, tVoid, tRoutine, tLabel, tInt, tRat, tStr
} Type;

char *gvsTypeStrings[] =
{
  "<null>", "void", "routine", "label", "int", "rat", "str"
};

typedef enum
{
  repNull, repIdent, repLiteral
} RepresentationType;

char *gvsRepresentationTypeStrings[] =
{
  "<null>", "identifier", "literal"
};

// Symbol //
typedef struct Symbol_tag
{
  RepresentationType rep;
  Type type;
  bool constant;

  void *value; // the value held in the symbol:
               // TYPE         VALUE          MEANING
               // ----         -----          -------
               // tRoutine     Routine*       routine
               // tLabel       int*           statement index
               // tInt         int*           integer
               // tRat         float*         rational
               // tStr         char*          string
               // tNull        (not defined)  (other)

  struct Symbol_tag *next;
  char *key;
} Symbol;

Symbol *sym_new (char *aKey);
Symbol *sym_copy (Symbol *symbol);
void sym_del (Symbol *symbol);
void sym_del_savevalue (Symbol *symbol);
void sym_set (Symbol *symbol, RepresentationType aRep, Type aType, bool aConstant, void *aValue);

// Parameter //
typedef struct Parameter_tag
{
  Type type;
} Parameter;

Parameter *parameter_new (Type t);
void parameter_del (Parameter *p);

// LocalVar //
typedef struct LocalVar_tag
{
  Symbol *symbol;
  struct LocalVar_tag *next;
  struct LocalVar_tag *prev;
} LocalVar;

// Routine //
typedef struct Routine_tag
{
  // External Use //
  Type returnType;
  int location;
  bool returnedFrom;
  int numOfParameters; // number of parameters
  Parameter *parameters[ROUTINE_MAX_PARAMETERS]; // an array of all parameters

  // Internal Use Only //
  LocalVar *lvHead; // a doubly-linked list holding all variables declared
  LocalVar *lvTail; //  ... in this routine; used for recursion
  LocalVar *lvCurr;
   
} Routine;

Routine *routine_new ();
void routine_del (Routine *routine);
void routine_addLocalVar (Routine *aRoutine, Symbol *aSymbol);
void routine_addParameter (Routine *aRoutine, Type aType);

// Bucket //
typedef struct
{
  Symbol *head;
  Symbol *tail;
} Bucket;

void buc_init (Bucket *bucket);
Symbol *buc_search (Bucket *bucket, char *aKey);
Symbol *buc_enter (Bucket *bucket, char *aKey);

// Table //
typedef struct
{
  Bucket **array;     // hash array
} Table;

Table *tab_new ();
void tab_del (Table *table);
Symbol *tab_search (Table *table, char *aKey);
Symbol *tab_enter (Table *table, char *aKey);

// Hash //
int hash (char *as);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

Parameter *parameter_new (Type t)
{
  Parameter *temp = (Parameter *) JALLOC(sizeof(Parameter));
  temp->type = t;
  return temp;
}

void parameter_del (Parameter *p)
{
  JFREE(p);
}

Routine *routine_new ()
{
  Routine *temp;
  int i;

  temp = (Routine *) JALLOC(sizeof (Routine));
  temp->returnType = tNull;
  temp->location = 0;
  temp->returnedFrom = false;

  temp->numOfParameters = 0;
  for (i = 0; i < ROUTINE_MAX_PARAMETERS; i++)
  {
    temp->parameters[i] = NULL;
  }
  temp->lvHead = NULL;
  temp->lvTail = NULL;
  temp->lvCurr = NULL;

  return temp;
}

void routine_del (Routine *routine)
{
  LocalVar *temp;
  int i;
   
  // free all the Parameters //
  for (i = 0; i < routine->numOfParameters; i++)
  {
    parameter_del (routine->parameters[i]);
  }

  // free all the LocalVars //
  while (routine->lvHead != NULL)
  {
    temp = routine->lvHead->next;
    JFREE(routine->lvHead);
    routine->lvHead = temp;
  }

  if (routine) JFREE(routine);  
}

//// routine_addLocalVar
//// Add a local variable to the list of local variables declared
////   in the routine
//// Created 0221, 18 Jan 2003
void routine_addLocalVar (Routine *aRoutine, Symbol *aSymbol)
{
  LocalVar *lv;

  // Ensure that the variable is not added more than once //
  lv = aRoutine->lvHead;
  while (lv != NULL)
  {
    if (lv->symbol == aSymbol)
    {
      err (true, EF_SYM_READDED_TO_LOCVAR_LIST, aSymbol->key);
      erre (EFF_TERMINATE);
    }
    lv = lv->next;
  }

  // Put the new LocalVar at the head of the list //
  lv = (LocalVar *) JALLOC(sizeof (LocalVar));
  lv->symbol = aSymbol;
  lv->next = aRoutine->lvHead;
  lv->prev = NULL;
  if (aRoutine->lvHead != NULL) aRoutine->lvHead->prev = lv;
  aRoutine->lvHead = lv;
  if (aRoutine->lvTail == NULL) aRoutine->lvTail = lv;
}

void routine_setCurrLocalVarToTail (Routine *aRoutine)
{
  aRoutine->lvCurr = aRoutine->lvTail;
}

void routine_setCurrLocalVarToHead (Routine *aRoutine)
{
  aRoutine->lvCurr = aRoutine->lvHead;
}

Symbol *routine_getPrevLocalVar (Routine *aRoutine)
{
  Symbol *temp;

  if (aRoutine->lvCurr)
  {
    temp = aRoutine->lvCurr->symbol;
    aRoutine->lvCurr = aRoutine->lvCurr->prev;
    return temp;
  }
  
  return NULL;
}

Symbol *routine_getNextLocalVar (Routine *aRoutine)
{
  Symbol *temp;

  if (aRoutine->lvCurr) 
  {
    temp = aRoutine->lvCurr->symbol;
    aRoutine->lvCurr = aRoutine->lvCurr->next;
    return temp;
  }

  return NULL;
}

//// routine_addParameter
//// Add a parameter to the Routine, this is for argument passing
//// Created: 0221, 18 Jan 2003
void routine_addParameter (Routine *aRoutine, Type aType)
{
  aRoutine->parameters[aRoutine->numOfParameters] = parameter_new (aType);
  aRoutine->numOfParameters++;
}

//// sym_new
Symbol *sym_new (char *aKey)
{
  Symbol *temp = (Symbol *) JALLOC(sizeof (Symbol));

  temp->rep = repNull;
  temp->type = tNull;
  temp->constant = false;
  temp->value = NULL;

  temp->next = NULL;
  temp->key = k_clone(aKey);

  return temp;
}

//// sym_copy
//// Makes a copy of the symbol and returns it. Does not copy
////   the "next" part.
//// Created: 0157 11/27/02
Symbol *sym_copy (Symbol *symbol)
{
  Symbol *temp = (Symbol *) JALLOC(sizeof (Symbol));

  temp->rep = symbol->rep;
  temp->type = symbol->type;
  temp->constant = symbol->constant;

  if (temp->type == tRoutine)
  {
    err (true, EF_CANNOT_CLONE_A_ROUTINE, NULL);
    erre (EFF_TERMINATE);
  }
  else if ((temp->type == tLabel) || (temp->type == tInt))
  {
    temp->value = k_new_n (*(int *)(symbol->value));
  }
  else if (temp->type == tRat)
  {
    temp->value = k_new_f (*(float *)(symbol->value));
  }
  else if (temp->type == tStr)
  {
    temp->value = k_clone ((char *)(symbol->value));
  }
  else
  {
    temp->value = NULL;
  }
  
  temp->key = k_clone (symbol->key);

  return temp;
}

//// sym_del
void sym_del (Symbol *symbol)
{
  // symbol may be NULL //
  if (!symbol) return;

  if (symbol->value)
  {
    if (symbol->type == tRoutine)
    {
      routine_del (symbol->value);
    }
    else
    {
      JFREE(symbol->value);
    }
  }
  k_del (symbol->key);
  JFREE(symbol);
}

//// sym_del_savevalue
//// Deletes the Symbol, but does not delete the value
void sym_del_savevalue (Symbol *symbol)
{
  // symbol may be NULL //
  if (!symbol) return;

  k_del (symbol->key);
  JFREE(symbol);
}

//// sym_set
void sym_set (Symbol *symbol, RepresentationType aRep, Type aType, bool aConstant, void *aValue)
{
  symbol->rep = aRep;
  symbol->type = aType;
  symbol->constant = aConstant;

  // given value is NULL, so set the value as the default for each type //
  if (aValue == NULL)
  {
    // determine the type of the declaration //
    if (aType == tNull) symbol->value = NULL;
    else if (aType == tLabel) symbol->value = k_new_n (0);
    else if (aType == tRoutine) symbol->value = routine_new (tNull, 0);
    else if (aType == tInt) symbol->value = k_new_n (0);
    else if (aType == tRat) symbol->value = k_new_f (0.0);
    else if (aType == tStr) symbol->value = k_clone ("");
  }
  else
  {
    if (aType == tNull) symbol->value = NULL;
    else if (aType == tLabel) symbol->value = k_new_n (*(int *)aValue);
    else if (aType == tRoutine)
    {
      err (true, EF_CANNOT_CLONE_A_ROUTINE, NULL);
      erre (EFF_TERMINATE);
    }
    else if (aType == tInt) symbol->value = k_new_n (*(int *)aValue);
    else if (aType == tRat) symbol->value = k_new_f (*(float *)aValue);
    else if (aType == tStr) symbol->value = k_clone (aValue);
  }
}

//// buc_init
void buc_init (Bucket *bucket)
{
  bucket->head = NULL;
  bucket->tail = NULL;
}

//// buc_search
//// Returns pointer to the Symbol with the matching
//// symbol key, or NULL if not found
Symbol *buc_search (Bucket *bucket, char *aKey)
{
  Symbol *temp = bucket->head;

  // Loop to find the matching symbol key //
  while (temp)
  {
    if (k_same (temp->key, aKey))
    {
      return temp;
    }
    else
    {
      temp = temp->next;
    }
  }

  // Not found; return NULL //
  return NULL;
}

//// buc_enter
//// Returns pointer to the Symbol; If not found, it will be created
Symbol *buc_enter (Bucket *bucket, char *aKey)
{
  Symbol *temp;

  temp = buc_search (bucket, aKey);

  // Not found; create and enter it //
  if (temp == NULL)
  {
    temp = sym_new (aKey);

    // First element //
    if (bucket->head == NULL)
    {
      bucket->head = bucket->tail = temp;
    }
    // Other elements //
    else
    {
      bucket->tail->next = temp;
      bucket->tail = temp;
    }
  }

  // Found //
  else
  {
  }

  // Return //
  return temp;
}

//// tab_new
Table *tab_new ()
{
  int i;
  Table *temp = (Table *) JALLOC(sizeof (Table));

  // Initialize the hash array to NULL //
  temp->array = (Bucket **) JALLOC(sizeof (Bucket *) * TABLE_HASH_ARRAY_SIZE);
  for (i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    temp->array[i] = NULL;
  }

  return temp;
}

//// tab_del
void tab_del (Table *table)
{
  int i;

  // free each element in array //
  for (i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    JFREE(table->array[i]);
  }
  JFREE(table->array);

  JFREE(table);
}

//// tab_search
//// Returns pointer to the node, or NULL if not found
Symbol *tab_search (Table *table, char *aKey)
{
  Bucket *pBucket;
  Symbol *pSymbol;

  pBucket = table->array[hash (aKey)];

  // There is no Bucket at that hash; Not found //
  if (pBucket == NULL)
  {
    return NULL;
  }
  // Bucket found at that hash //
  else
  {
    pSymbol = buc_search (pBucket, aKey);
    return pSymbol;
  }
}

//// tab_enter
//// Returns pointer to the node; If not found, it will be created
Symbol *tab_enter (Table *table, char *aKey)
{
  int nHash;
  Bucket *pBucket;
  Symbol *pSymbol;

  nHash = hash (aKey);
  pBucket = table->array[nHash];

  // There is no Bucket at that hash; Create Bucket //
  if (pBucket == NULL)
  {
    pBucket = (Bucket *) JALLOC(sizeof (Bucket));
    buc_init (pBucket);
    table->array[nHash] = pBucket;
  }

  // Create Symbol //
  pSymbol = buc_enter (pBucket, aKey);

  return pSymbol;
}

//// hash
//// hash function for the "Alpha" algorithm (Aho p435)
int hash (char *as)
{
  int i;
  int len;
  int total;

  len = strlen (as);
  total = 0;

  for (i = 0; i < len; i++)
  {
    total += as[i] * TABLE_HASH_ALPHA;
  }

  return (total % TABLE_HASH_ARRAY_SIZE);
}

#endif
